''' A module to modify FITS file images.

Written by Azin Khan. Summer 2012.'''



import numpy

import pyfits

import scipy

import ndimage

import time

import numpy.ma as ma







def offset(images, offsets, integer=False):

    '''

    This function takes a list of the images and arranges them separately into

    a larger array such that the correct position of the image is achieved 

    using the offsets given.

    

    Input:

        images = List of images to be arranged. Even if only one image is to be

                 offset, the data-type of this parameter must still be a list

                 e.g. images = [some_array], not images = some_array.

        

        offsets = A tuple of lists documenting the correct offset for each 

        		  image. These must be in x,y) form



        integer = Input to determine whether integer or non-integers offsets 

        		  are used. If False, then bilinear interpolation is used.

                  Non-integer offsets are set as default.

        

    Output:

        Returns a list of the images correctly positioned on a larger image,

        ready to be combined.



    Note:

        The list of images and the list of offsets must correspond; the image

        from which the offsetsare calculated must have an offset entry 

        of (0,0).



        The larger arrays size is calculated using the offset file.

        This ensures that the larger array is of minimum size to reduce 

        filesize and inefficiencies, but will fit all of the required images.



        If non-integer offsets are given with integer == True, then the offsets

        are rounded to the nearest integer.



    '''



    X_offsets = offsets[0] 

    Y_offsets = offsets[1]



    if type(images) is not list:

        raise TypeError, 'The images must be given in a list'



    else:

        s = images[0].shape

        w = s[0]

        y = s[1]

        x = s[2]



        

    # Determine the required size of the composite image.

    if max(X_offsets) <0:

        X_dimension = x + numpy.absolute(min(X_offsets))

        x_pos = numpy.absolute(min(X_offsets))

    elif min(X_offsets) >0:

        X_dimension = x + max(X_offsets)

        x_pos = 0

    else:

        X_dimension = x + max(X_offsets) + numpy.absolute(min(X_offsets))

        x_pos = numpy.absolute(min(X_offsets))

        

    if max(Y_offsets) <0:

        Y_dimension = y + numpy.absolute(min(Y_offsets))

        y_pos = numpy.absolute(min(Y_offsets))

    elif min(Y_offsets) >0:

        Y_dimension = y + max(Y_offsets)

        y_pos = 0

    else:

        Y_dimension = y + max(Y_offsets) + numpy.absolute(min(Y_offsets))

        y_pos = numpy.absolute(min(Y_offsets))



    final_list = []

    

    if integer:

        # Create a larger array of the size required and place the

        # image given at the correct location within the larger array.

        for i, image in enumerate(images):

            # Round the offsets if integer offsetting is required.

            large = numpy.ones((w, round(Y_dimension), round(X_dimension)))

            x_offset = round(X_offsets[i])

            y_offset = round(Y_offsets[i])

            # Mask large array so only the portion with the image in it is valid.

            large[0:w, y_pos+y_offset:y_pos+y_offset+y, x_pos+x_offset:x_pos+

                  x_offset+x] = 0

            masked_large = ma.masked_equal(large,1)

            masked_large[0:w, y_pos+y_offset:y_pos+y_offset+y, x_pos

                                         +x_offset:x_pos+x_offset+x] += image

            

            final_list.append(masked_large)

    else:

        for i, image in enumerate(images):

            # Use bi-linear interpolation to offset the images correctly.

            large = numpy.ones((w, Y_dimension, X_dimension))

            x_offset = X_offsets[i]

            y_offset = Y_offsets[i]

            # Determine the 'non-integer' part of the offset by truncating it.

            # and then interpolate.

            x_shift = int(x_offset)-x_offset

            y_shift = int(y_offset)-y_offset

            shift = [0,y_shift,x_shift]

            interp_image = ndimage.interpolation.shift(image, shift, order=1)

            # Now the remaining offset is an integer and can be simply added on.

            # Use the same procedure as above.

            large[0:w, y_pos+y_offset:y_pos+y_offset+y, x_pos+x_offset:x_pos+

                  x_offset+x] = 0

            masked_large = ma.masked_equal(large,1)

            masked_large[0:w, y_pos+y_offset:y_pos+y_offset+y, x_pos+

                                             x_offset:x_pos+x_offset+x] += image

            

            final_list.append(masked_large)

         

    return final_list

            

def precise_combine(images, bpms, threshold = 0, fill=0):

    '''

    This function combines a list of given images and any bad pixel masks into

    one image.



    Input:

        images = List of images to be combined.

        

        bpms = List of corresponding bad pixel mask arrays.

        

        threshold = Threshold values for bad pixels.

        

        fill = Fill value for masked pixels (0 by default to ensure that they

               are properly ignored).



        mask = Determines if the images are to be masked. If set to True, then

               the images will be masked according to the BPM information.

        

    Output:

        Returns the composite of the images.

        

    Note:

        The procedure is to take a wavelength 'slice' across all the images and

        then average using ma.average to ignore masked entries. This results in

        a final slice that is appended to a list of slices and eventually

        converted into a numpy array once all slices have been added to the

        list.This is a rather slow way to combine datacubes but fully accounts

        for the bad pixel masks, resulting in the smoothest image possible



        It is advisable to set mask = True and input an unmasked array as

        otherwise the ma.count function may not operate correctly. This appears

        to be a problem inherent in the numpy.ma module (see numpy.ma

        documentation).



    '''



    final = []

    bpm_info = []

    wavelengths = images[0].shape[0]





    for k in range(wavelengths):

        mini = []

        minierror = []

        for i, image in enumerate(images):

            # For each wavelength, find the corresponding slice in each

            # image and create a small data cube with this information.

            # This small data cube has a third dimension of image index.

            mini.append(image[k])

            if bpms:

                minierror.append(bpms[i][k])

                    

        # Convert to an array once all the slices have been found.

        error = ma.array(minierror)

        minicube = ma.array(mini)

 

        # Mask the array then average over it, hence combining the

        # different images at this wavelength into one final slice.

        if bpms:

            masked_mini = ma.masked_where(error>threshold, minicube, copy=False)

        else:

            masked_mini = minicube

        bpm_infoslice = ma.count(masked_mini, axis=0)

        bpm_info.append(bpm_infoslice)

        averaged_slice = ma.average(masked_mini, axis=0)

        final.append(numpy.array(averaged_slice))

            

    # Convert the results into numpy arrays.

    # Pyfits cannot handle ma.array (yet).

    # Any remaining masked pixels are set to zero in the process.

    final_array = numpy.array(final)

    bpm_final_info = numpy.array(bpm_info)



    return final_array, bpm_final_info





def open_refine(image_stem, error_stem, number, X_cut=0, Y_cut=0):

    '''

    This function opens a FITS file and extracts the data as a numpy array.

    The refine element is included to remove edge pixels from the outset;

    they are simply cropped out.

        

    Input:

        image_stem = The filname stem for the image.



        error_stem = The filename stem for the bpm file.



        number = The number of the file to be opened.



        X_cut, Y_cut = The number of pixels to be sliced from each edge.

                        This is done symmetrically



        Example input:



            open_refine('ms0%d.sci.cube.fits', 'ms0%d.bpm.cube.fits', 83, 5,10)



            This would open ms083.sci.cube.fits as the image and

            ms083.bpm.cube.fitsas the bpm array with 5 pixels cropped

            from each end of thex axis and 10 pixels cropped from each end

            of the y axis.Note that if the filenames are numbered with 0 as

            the leading digit, then this must be included in the stem, not

            the number.



    Output:

        Returns a list of two arrays: [Image, BPM].



    '''

    

    image = pyfits.open(image_stem %number)

    image_data = refine_array(image[0].data,X_cut,Y_cut)

    error = pyfits.open(error_stem %number)

    error_data = refine_array(error[0].data,X_cut,Y_cut)



    

    return [image_data, error_data]



def refine_array(data, x_cut, y_cut):

    'Simple function to crop an array symmetrically'

    size = data.shape

    w = size[0]

    y = size[1]

    x = size[2]

    refined_data = data[0:w, y_cut:y-y_cut, x_cut:x-x_cut]

    return refined_data



def background_remove(A1, A2):

    '''

    This function performs a pairwise background subtraction on two arrays



    '''

    B1 = A1 - A2

    B2 = A2 - A1



    return B1, B2



if __name__ == '__main__':

    

    offsets = ([0, 26.54, 27.65, -20.99], [0, 2.589, 8.539, 6.959])

    threshold = 0.5

    newfilename = input('Enter New Filename as a string.'

                        'This must have the extension .fits\n')



    arrays = []

    errors = []

    start = time.time()

    for i in range(83,87,1):

        data = open_refine('ms0%d.sci.cube.fits', 'ms0%d.bpm.cube.fits',i,5,5)

        arrays.append(data[0])

        errors.append(data[1])

        

    bkgrnd = []

    for k in range(0,len(arrays),2):

        bk = background_remove(arrays[k], arrays[k+1])

        bkgrnd.append(bk[0])

        bkgrnd.append(bk[1])

        

    offset_images = offset(bkgrnd, offsets, integer=False)

    offset_bpm = offset(errors, offsets, integer=False)



        

    combined = precise_combine(offset_images)

    cig = ndimage.gaussian_filter(combined, sigma =3)

    hdu = pyfits.PrimaryHDU(cig)

    hdulist = pyfits.HDUList([hdu])

    hdulist.writeto(newfilename)

    print (time.time()-start), 'seconds'    

        



        







        

    





        

    

